use ExtUtils::MakeMaker;
use ExtUtils::Liblist;
use Config;

# See lib/ExtUtils/MakeMaker.pm for details of how to influence
# the contents of the Makefile that is written.

# ExtUtils::Liblist is broken
@findlib::ISA = qw(ExtUtils::Liblist);
sub findlib::lsdir { shift; grep /$_[0]/, <$_[1]>}
sub findlib::file_name_is_absolute { 1 } # Used for warnings only...

 
# Supported windowing interfaces
our $is_valid_interface =
{
  'FREE_GLUT'	=> 'FreeGLUT, preferred over GLUT',
  'GLUT'	=> 'GLUT; this really FreeGLUT on Linux',
  'GLX'		=> 'X11; requires a running XServer',
  'AGL'		=> 'GLUT framework on Mac OS X',
  'W32API'	=> 'Uses GLUT32 on CYGWIN, instead of GLX'
};


# Makefile.PL Usage
if ($ARGV[0] =~ m|^([-/]*)?h(elp)?|i)
{
  print qq
  {
    USAGE: perl Makefile.PL [OPTIONS]


    OPTIONS include:

    help                This help message

    verbose             Display additional status info

    dist=NO_EXCLUSIONS  Build with no OpenGL Extension exclusions
    dist=EXCLUSIONS     Override NO_EXCLUSIONS default for Win32

    interface=XFACE     Build for a specific windowing interface
                        Currently supports:
  };

  print "\n";
  foreach $xface (keys %$is_valid_interface)
  {
    print substr("      $xface                  ",0,24).
      "$is_valid_interface->{$xface}\n";
  }
  print "\n";

  exit(0);
}


# Get debugging flags
our $verbose = 0;
if ( grep { m/^verbose/i } @ARGV )
{
  $verbose = 1;
  # Strip out interface args
  @ARGV = grep { !m/^verbose=/i } @ARGV;
}


# Get distribution build flags
our $dist_flags = {};
if ( grep { m/^dist=/i } @ARGV )
{
  foreach my $arg (@ARGV)
  {
    $dist_flags->{uc($1)}++ if ($arg =~ m/^dist=(\w+)/i);
  }

  # Setting both EXCLUSIONS and NO_EXCLUSIONS cancels each other out
  if ($dist_flags->{NO_EXCLUSIONS} && $dist_flags->{EXCLUSIONS})
  {
    $dist_flags->{NO_EXCLUSIONS} = undef;
    $dist_flags->{EXCLUSIONS} = undef;
  }

  # Strip out interface args
  @ARGV = grep { !m/^dist=/i } @ARGV;
}
# Default to NO_EXCLUSIONS on Windows
if ($^O eq 'MSWin32')
{
  $dist_flags->{NO_EXCLUSIONS} = $dist_flags->{EXCLUSIONS} ? 0 : 1;
}


# Detect CYGWIN
open(SAVEERR, ">&STDERR");
close(STDERR);
my $uname = `uname`;
open(STDERR, ">&SAVEERR");
our $IS_CYGWIN = ($uname =~ m|^CYGWIN|);


# Look for available libs
our @libdirs = qw
{
  -L/usr/lib
  -L/usr/X11R6/lib
  -L/usr/local/lib
  -L/usr/openwin/lib
  -L/usr/lib/xorg/modules/extensions
  -L/usr/X11R6/lib/modules/extensions
};

our @check_libs = qw
{
  -lGL
  -lGLU
  -lglut
  -lfreeglut
  -lopengl
  -lGLX
  -lglx
  -lglitz-glx
  -lglu
  -lGLUT
  -lOSMesa
  -lMesaGL
  -lMesaGLU
  -lMesaGLX
  -lMesaGLUT
};

our $found_libs = get_libs(@libdirs,@check_libs);
die "No OpenGL libs found\n" if (!scalar($found_libs));


# Check for OpenGL installation
our($DEFS,$glut_lib,$glut_def) = get_extensions($found_libs,
  $dist_flags->{NO_EXCLUSIONS});


# Get user-specified interface
my $interface_lib;
if ( grep { m/^interface=/i } @ARGV )
{
  my @my_argv = @ARGV;    # safe copy
  my @interface_opts = ();

  foreach my $arg (@my_argv)
  {
    if ($arg =~ m/^interface=(\w+)/i)
    {
      my $interface = uc($1);
      my $valid_type = $is_valid_interface->{$interface};
      next if (!$valid_type);

      # test if requested interfaces supported
      if ($interface eq 'W32API')
      {
        next if (!$found_libs->{GLUT32});
        #$found_libs->{GL} = $found_libs->{OPENGL32};
        #$found_libs->{GLU} = $found_libs->{GLU32};
        $found_libs->{GLUT} = $found_libs->{GLUT32};
        $found_libs->{GLX} = undef;
      }
      elsif ($interface eq 'AGL')
      {
        next if ($^O ne 'darwin' || !$found_libs->{FREE_GLUT});
      }
      elsif ($interface eq 'GLX')
      {
        next if (!$found_libs->{GL});
      }
      elsif (!$found_libs->{$interface})
      {
        next;
      }

      push(@interface_opts,$interface);
    }
  }
  print "Got the following valid interface options: @interface_opts\n" if $verbose;

  if (scalar(@interface_opts) > 1)
  {
    die "ERROR: multiple opengl interfaces requested: @interface_opts, quitting.\n";
  }

  ($interface_lib) = @interface_opts;
  print "User set interface=$interface_lib\n" if $verbose;

  # In case someone wants to force the use of the original GLUT API
  if ($interface_lib eq 'GLUT')
  {
    $found_libs->{FREE_GLUT} = undef;
  }

  # Strip out interface args
  @ARGV = grep { !m/^interface=/i } @ARGV;
}
# Otherwise use available defaults
elsif ($IS_CYGWIN)
{
  if ($found_libs->{GLX})
  {
    $interface_lib = 'GLX'; # CYGWIN will use X11 interface by default
  }
  elsif($found_libs->{GLUT32})
  {
    $interface_lib = 'W32API';
  }
  else
  {
    die "No suitable interface found for CYGWIN\n";
  }
}
elsif ($^O eq 'darwin' && $found_libs->{FREE_GLUT})
{
  $interface_lib = 'AGL';
}
elsif ($found_libs->{FREE_GLUT})
{
  $interface_lib = 'FREE_GLUT';
}
elsif ($found_libs->{GLUT})
{
  $interface_lib = 'GLUT';
}
elsif ($^O eq 'MSWin32')
{
  die "Windows requires FreeGLUT or GLUT\n";
}
else
{
  $interface_lib = 'GLX'; # otherwise default to GLX
}
print "Using interface: $interface_lib\n";


# Marshall compiler defines
foreach my $key (keys %$found_libs)
{
  next if (!$found_libs->{$key});
  $DEFS .= " -DHAVE_$key";
}


# Configure MakeMaker
my($INCS,$LIBS,$OPTS);
my $EXES = [];
my $DYNS = {};
if (($^O eq 'MSWin32') && !$IS_CYGWIN)		# Win32
{
  $LIBS = '-lopengl32 -lglu32 -l';
  $LIBS .= ($interface_lib eq 'GLUT') ? 'glut32' : 'FreeGLUT/freeglut';

  $EXES = ['FreeGLUT/freeglut.dll'];
  $OPTS = "/O2gity /GDs";
}
elsif ($interface_lib eq 'W32API')		# CYGWIN
{
  $LIBS = "-L/usr/lib/w32api -lopengl32 -lglu32 -lglut32";
  $INCS = '-I/usr/include/w32api';

  $DEFS .= " -DHAVE_W32API";
}
elsif ($interface_lib eq 'AGL')			# Mac OS X
{
  $DYNS = { OTHERLDFLAGS => "-framework OpenGL -framework GLU -framework GLUT" };
}
else # Everyone else
{
  # Handle obfuscated GLX
  sub cfile_text
  {
    my($filename) = @_;

    # Use $Config{cpprun}, instead of $Config{cpp}. cpp is simply set to
    # 'cpp' after Configure has run, which is not useful, since it lives
    # in /lib/ on many systems, which is normally not on the path.  This
    # is documented, but silly behaviour.  $Config{cpprun} is what is
    # documented to be set to a cpp that will take stuff on stdin

    my $cmd = "echo '\#include \"$filename\"' | ".
      "$Config{cpprun} $includes $Config{cppflags} $Config{cppminus} 2>&1";

    return `$cmd`;
  }

  # quite often GLX is in -lGL...  Test for GL/glx.h instead...
  my $out = cfile_text('GL/glx.h');

  # The cpp would not let this macro through. Check for something else
  # that still exists after the cpp pass. a typedef, or type would work
  my $has_glx = ($out =~ m|GLXContext|);

  if ($has_glx)
  {
    if (!$found_libs->{GLX})
    {
      $DEFS .= " -DHAVE_GLX";

      # Handle OS2 - why?
      if ($^O eq 'os2')
      {
        $DEFS .= ' -DOS2_GL_DISABLE_MISSING_EXT -DAPIENTRY= -D__PM__';
      }
    }
    else
    {
      delete($found_libs->{GLX});
    }
  }
  elsif (!$found_libs->{GLX})
  {
    print STDERR "#\n### GLX not found (neither library, nor headers).\n#\n";
  }


  # Marshall libs
  my $libs = ' -l'.join(' -l',values(%$found_libs));
  my @more_libs = qw
  {
    -lXext
    -lXmu
    -lXi
    -lICE
    -lX11
    -lm
  };
  #-lXIE
  $LIBS = "@libdirs $libs @more_libs";

  my @includes = qw
  {
    -I/usr/include
    -I/usr/X11R6/include
    -I/usr/local/include
    -I/usr/openwin/include
  };
  $INCS = "@includes";

  $DEFS .= " -DGL_GLEXT_LEGACY";


  # Handle obfuscated gcc
  my $no_lgcc = (grep /^NO_LGCC$/, @ARGV);
  @ARGV = grep !/^NO_LGCC$/, @ARGV;

  if (!$no_lgcc)
  {
    # This may be a gcc compiler in disguise.
    # Should check for gccversion as well.
    if ($Config{cc} ne 'gcc' && !$Config{gccversion})
    {
      my $gcclibdir;
      my $gccout = `gcc -v 2>&1`;
      if ($gccout =~ /specs from (.*)/)	# gcc present
      {
        ($gcclibdir = $1) =~ s,\bspecs\s*$,,;
        $LIBS .= " -L$gcclibdir -lgcc";

        print STDERR qq
        {
#
### Non-gcc compiler, and gcc is present.
### Adding -lgcc as a dependency: your OpenGL libraries may be gcc-compiled.
### Use NO_LGCC command-line option to disable this.
#
        };
      }
    }
  }
}

WriteMakefile
(
  'NAME'	=> 'OpenGL',
  'VERSION_FROM'=> 'OpenGL.pm',
  'PM' 		=> { 'OpenGL.pm' => '$(INST_LIBDIR)/OpenGL.pm' },
  'AUTHOR'	=> "Bob 'grafman' Free ".'<grafman at graphcomp.com>',
  'OBJECT' 	=> '$(BASEEXT)$(OBJ_EXT) gl_util$(OBJ_EXT)',
  'XSPROTOARG'	=> '-noprototypes',
  'DEFINE'	=> $DEFS,
  'INC'		=> $INCS,
  'LIBS' 	=> $LIBS,
  'dynamic_lib'	=> $DYNS,
  'EXE_FILES'	=> $EXES,
  'OPTIMIZE'	=> $OPTS
);

exit 0;




# Find OpenGL-related Libs
sub get_libs
{
  my @glx_libs = @_;
  my $found = {};
  my $libs;

  # Win32
  if (($^O eq 'MSWin32') && !$IS_CYGWIN)
  {
    my $paths = $ENV{PATH};
    $paths =~ s|\\|/|g;
    my @paths = split(';\s*',$paths);

    $libs = {GL=>'opengl32.dll',GLU=>'glu32.dll',GLUT=>'glut32.dll',FREE_GLUT=>'freeglut.dll'};
    foreach my $key (keys %$libs)
    {
      foreach my $path (@paths)
      {
        $path =~ s|/$||;
        my $lib = $libs->{$key};
        next if (!-e "$path/$lib");
	# print "$path/$lib\n" if $verbose;
        $lib =~ s/\.dll$//i;
        $found->{$key} = $lib;
        last;
      }
    }

    # Install FreeGLUT if not found
    if (!$found->{FREE_GLUT})
    {
      my $dll_dir = $Config{installbin};
      #my $dll_dir = `echo %SystemRoot%\\system32`;
      #$dll_dir =~ s|[\n\r]||g;
      my $exec = "copy FreeGLUT\\freeglut.dll $dll_dir";
      `$exec`;

      if (-e "$dll_dir/freeglut.dll")
      {
        print "Installed freeglut.dll at $dll_dir\n";
        $found->{FREE_GLUT} = 'freeglut' 
      }
      else
      {
        print "Unable to install freeglut.dll to $dll_dir\n";
        print "Try manually copying FreeGLUT/freeglut.dll to your system folder.\n";
      }
    }
    return $found;
  }


  # Mac OS X
  if ($^O eq 'darwin' && -d '/System/Library/Frameworks/OpenGL.framework')
  {
    $found->{GL} = 'OpenGL.framework';
    $found->{GLU} = 'AGL.framework';
    $found->{FREE_GLUT} = 'GLUT.framework';
    $found->{GLUT} = $found->{FREE_GLUT};
    return $found;
  }


  # CYGWIN
  if ($IS_CYGWIN)
  {
    open(SAVEERR, ">&STDERR");
    close(STDERR);
    ($libs) = findlib->ext('-L/usr/lib/w32api -lopengl32 -lglu32 -lglut32',0);
    open(STDERR, ">&SAVEERR");

    foreach my $lib (split(' ',$libs))
    {
      next if ($lib =~ m|^-L|);
      $lib =~ s/^-l//;
      $found->{uc($lib)} = $lib;
    }
  }


  # Everyone else
  open(SAVEERR, ">&STDERR");
  close(STDERR);
  ($libs) = findlib->ext(join(' ',@glx_libs),0);
  open(STDERR, ">&SAVEERR");

  foreach my $lib (split(' ',$libs))
  {
    next if ($lib =~ m|^-L|);
    $lib =~ s/^-l//;

    if ($lib =~ m|([\w]*freeglut[\w]*)|i)
    {
      $found->{FREE_GLUT} = $1;
    }
    elsif ($lib =~ m|([\w]*glut[\w]*)$|i)
    {
      $found->{GLUT} = $1;
    }
    elsif ($lib =~ m|([\w]*glu[\w]*)$|i)
    {
      $found->{GLU} = $1;
    }
    elsif ($lib =~ m|([\w-]*glx)$|i)
    {
      $found->{GLX} = $1;
    }
    elsif ($lib =~ m|([\w]*gl)$|i)
    {
      $found->{GL} = $1;
    }

    if ($lib =~ m|([\w]*mesa(gl)?.*)|i)
    {
      $found->{MESA} = $1;
    }
  }


  if ($verbose)
  {
    print "libs:\n";
    foreach my $key (keys %{$found})
    {
      print "    $key = '$found->{$key}'\n";
    }
  }

  return $found;
}


# Test for available OpenGL Extensions
sub get_extensions
{
  my($found,$no_excl) = @_;
  print "Testing for OpenGL Extensions\n" if ($verbose);

  # clean up previous output files
  my $exc_file = 'gl_exclude.h';
  unlink($exc_file) if (-e $exc_file);
  my $glv_file = 'utils/glversion.txt';
  unlink($glv_file) if (-e $glv_file);


  # Only run the rest if GLUT is installed
  return if (!$found->{FREE_GLUT} && !$found->{GLUT} && !$found->{GLUT32});
  print "Testing GLUT version\n" if ($verbose);

  my $lib;
  my $def;
  if ($found->{FREE_GLUT})
  {
    $lib = $found->{FREE_GLUT};
    $def = "HAVE_FREE_GLUT";
  }
  elsif ($found->{GLUT32})
  {
    $lib = $found->{GLUT32};
    $def = "HAVE_W32API";
  }
  else
  {
    $lib = $found->{GLUT};
    $def = "HAVE_GLUT";
  }

  my $make_ver;
  if ($IS_CYGWIN)
  {
    $make_ver = ";make -f Makefile.cygwin GLUT_LIB=$lib GLUT_DEF=$def";
    print "cygwin glversion: '$make_ver'\n" if $verbose;
  }
  elsif ($^O eq 'MSWin32')
  {
    $make_ver = '&nmake -f makefile.mak GLUT_DEF='.$def;
    print "MSWin32 glversion: '$make_ver'\n" if $verbose;
  }
  elsif ($^O eq 'darwin')
  {
    $make_ver = ";make -f Makefile.macosx GLUT_LIB=$lib GLUT_DEF=$def";
    print "MacOSX glversion: '$make_ver'\n" if $verbose;
  }
  else
  {
    $make_ver = ";make -f Makefile GLUT_LIB=$lib GLUT_DEF=$def";
    print "glversion: '$make_ver'\n" if $verbose;
  }
  my $stat = qx(cd utils $make_ver clean $make_ver);
  print "\n$stat\n\n" if ($verbose);
  return if (!-e $glv_file);


  die "Unable to open glext_procs.h\n" if (!open(GLEXT,"glext_procs.h"));
  my @lines = <GLEXT>;
  close(GLEXT);

  # Parse glext_procs.h file
  my $no_ext = {};
  foreach my $line (@lines)
  {
    next if ($line !~ m|\#ifndef NO_([^\s]+)|);
    my $ext = $1;
    next if ($ext =~ m|^GL_VERSION_|);
    $no_ext->{$ext}++;
  }


  # Parse glversion.txt file
  die "Unable to open $glv_file\n" if (!open(GLDATA,$glv_file));
  my $gldata = {};
  my @gldata = <GLDATA>;
  close(GLDATA);

  foreach my $line (@gldata)
  {
    $line =~ s|[\r\n]+||;
    my($key,$val) = split('=',$line);
    $gldata->{$key} = $val;
  }


  # Create gl_exclude.h
  die "Unable to write to $exc_file\n" if (!open(GLEXC,">$exc_file"));
  print GLEXC "// OpenGL Extension Exclusions - may be modified before building.\n";
  print GLEXC "//\n";
  print GLEXC "// Generated for ".$gldata->{VENDOR}.", ".$gldata->{RENDERER}."\n";
  print GLEXC "// OpenGL v".$gldata->{VERSION}.", using ";

  # Fix GLUT flags based on results
  if ($gldata->{GLUT} > 0)
  {
    print GLEXC 'FreeGLUT v'.$gldata->{GLUT}."\n";
    $def = 'HAVE_FREE_GLUT';

    if (!$found->{FREE_GLUT})
    {
      $found->{FREE_GLUT} = $lib;
    }
  }

  # Make an empty exclusion file if a Windows distribution build
  if ($no_excl)
  {
    print "OpenGL Extension exclusions disabled\n";
    print GLEXC "//\n";
    print GLEXC "// Exclusions omitted for distribution build.\n";
  }
  else
  {
    print GLEXC "\n";
    if ($gldata->{VERSION} =~ m|^(\d\.\d+)|)
    {
      my $ver = $1;
      print GLEXC "#define NO_GL_VERSION_2_0\n" if ($ver < 2.0);
      print GLEXC "#define NO_GL_VERSION_1_5\n" if ($ver < 1.5);
      print GLEXC "#define NO_GL_VERSION_1_4\n" if ($ver < 1.4);
      print GLEXC "#define NO_GL_VERSION_1_3\n" if ($ver < 1.3);
      print GLEXC "#define NO_GL_VERSION_1_2\n" if ($ver < 1.2);
      print GLEXC "#define NO_GL_VERSION_1_1\n" if ($ver < 1.1);
    }

    foreach my $ext (split(' ',$gldata->{EXTENSIONS}))
    {
      next if (!$no_ext->{$ext});
      $no_ext->{$ext} = 0;
    }

    foreach my $ext (sort keys(%$no_ext))
    {
      next if (!$no_ext->{$ext});
      print GLEXC "#define NO_$ext\n";
    }
  }

  close(GLEXC);
  #print "Have Version Data\n";

  return('-DHAVE_VER',$lib,$def);
}
